import importlib
import os
from typing import Optional

from config import get_config
from fastapi.routing import Mount
from fastapi.staticfiles import StaticFiles
from libs.api.api_handler import ApiHandler

config = get_config()


class Bootstrap:
    def __init__(self) -> None:
        self.controllers = []
        self.current_path: str = os.path.relpath(os.path.join(os.getcwd()))
        path = self.get_project_path()
        modules_path = self.get_module_path(path)
        for root, _, files in os.walk(modules_path):
            for file in files:
                if 'controller' in file.lower() and file.endswith('.py'):
                    module_path = (
                        root.replace('/', '.')[2:]
                        + '.'
                        + file.replace('.py', '')
                    )
                    module_path = '.'.join(module_path.split('.')[1:])
                    modules = importlib.import_module(module_path)
                    for d in dir(modules):
                        if 'controller' in d.lower():
                            obj = getattr(modules, d)
                            if hasattr(obj, 'router'):
                                self.controllers.append(obj())

    def get_project_path(self) -> Optional[str]:
        for root, dirs, _ in os.walk(self.current_path):
            if root.startswith('./.') or root.endswith('.'):
                continue
            if 'modules' in dirs:
                return root

    def get_module_path(self, path: str) -> Optional[str]:
        for root, dirs, files in os.walk(path):
            if (
                '__pycache__' in dirs
                or '__pycache__' in root
                or 'modules' not in root
            ):
                continue
            return root

    def setup_fastapi_app(self, app: ApiHandler):
        for controller in self.controllers:
            if hasattr(controller, 'static'):
                static = getattr(controller, 'static')
                app.mount(
                    f'/{static["name"]}',
                    StaticFiles(directory=static['directory']),
                    static['name'],
                )
            app.add_router(controller.router)
        app.mount(
            '/static', StaticFiles(directory='backend/static'), name='static'
        )
        paths_to_create = [
            config.PATH_SOURCE_VIDEOS,
            config.PATH_SOURCE_COVERS,
            config.PATH_PROCESSED_VIDEOS,
        ]
        for path in paths_to_create:
            if not os.path.isdir(path):
                os.makedirs(path)
        app.mount(
            '/cover',
            StaticFiles(directory=config.PATH_SOURCE_COVERS),
            name='cover',
        )



from typing import Optional

from libs.database import DatabaseHandler
from sqlalchemy import select

from .models import TodoModel
from .schemas import CreateTodoSchema, TodoUpdateSchema


class TodoRepository:

    database_handler = DatabaseHandler()

    def new_todo(self, user_id: int, todo_data: CreateTodoSchema) -> TodoModel:
        with self.database_handler as session:
            todo = TodoModel(user_id=user_id, **todo_data)
            session.add(todo)
            session.expire_on_commit = False
            session.commit()

    def get_todo(self, todo_id: int) -> Optional[TodoModel]:
        with self.database_handler as session:
            query = select(TodoModel.id == todo_id)
            todo = session.scalar(query)
            session.expire_on_commit = False
            session.commit()
            return todo

    def get_all_todos(self) -> list[TodoModel]:
        with self.database_handler as session:
            query = select(TodoModel)
            todo = session.scalars(query).all()
            session.expire_on_commit = False
            session.commit()
            return todo

    def delete_todo(self, todo_id: int) -> Optional[TodoModel]:
        with self.database_handler as session:
            query = select(TodoModel.id == todo_id)
            todo = session.scalar(query)

            session.expire_on_commit = False
            session.commit()
            return todo

    def update_todo(
        self, todo_id: int, todo_new_data: TodoUpdateSchema
    ) -> Optional[TodoModel]:
        with self.database_handler as session:
            query = select(TodoModel.id == todo_id)
            todo = session.scalar(query)
            todo = todo(**todo_new_data)
            session.add(todo)
            session.expire_on_commit = False
            session.commit()
            session.refresh(todo)
            return todo



from typing import Optional, Type

from config import get_config
from sqlalchemy import Engine, create_engine
from sqlalchemy.orm import Session, sessionmaker


class DatabaseHandler:

    instance: Optional['DatabaseHandler'] = None
    config = get_config()

    def __init__(self) -> None:
        self._engine = create_engine(self.config.SQLITE_DB, max_overflow=-1)
        self.session_manger = sessionmaker(self._engine)

    def __new__(cls: Type['DatabaseHandler']) -> 'DatabaseHandler':
        if cls.instance is None:
            cls.instance = super(DatabaseHandler, cls).__new__(cls)
        return cls.instance

    def get_engine(self) -> Engine:
        return self._engine

    def get_session(self) -> Session:
        return self.session_manger()

    def __enter__(self):
        return self.session_manger()

    def __exit__(self, exc_type, exc_value, traceback):
        ...



import os
import pathlib
import shutil

from mako.template import Template

from ...config import DatabaseConfig, ModuleConfig, UserConfig

def run(
    project_name: str,
    module_config: ModuleConfig,
    user_config: UserConfig,
    database_config: DatabaseConfig,
):
    current_dir = os.path.dirname(__file__)
    shutil.copytree(os.path.join(current_dir, 'libs/'), os.path.join(project_name, 'libs'))
    shutil.copytree(os.path.join(current_dir, 'utils/'), os.path.join(project_name, 'utils'))
    os.makedirs(os.path.join(project_name, module_config.module_path))
    shutil.copytree(os.path.join(current_dir, 'modules/user/'), os.path.join(project_name, module_config.module_path, 'user'))
    shutil.copy(os.path.join(current_dir, 'config.py'), os.path.join(project_name, 'config.py'))
    shutil.copy(os.path.join(current_dir, 'main.py'), os.path.join(project_name, 'main.py'))
    shutil.copy(os.path.join(current_dir, 'app.py'), os.path.join(project_name, 'app.py'))
    if not os.path.isfile(os.path.join(current_dir, '.env')):
        with open('.env', 'w') as file:
            file.write('')
    with open('.env', 'a') as file:
        template = Template(open(os.path.join(current_dir, '.env')).read())
        content = template.render(project_name=project_name)
        file.write(f'\n{content}')

def new_module():
    ...



import sys

from pick import pick

from .config import DatabaseConfig, ModuleConfig, UserConfig
from .structs.default.run import run


class BifeCreateProject:
    def run(self):
        if len(sys.argv) == 1:
            print('Please, tell me the name of your project!')
            sys.exit(1)
        title = 'Welcome to BIFE, do you want use database?'
        database = self.database_config()
        user = self.setup_user_module()
        modules = self.setup_modules()
        run(sys.argv[3], modules, user, database)

    def database_config(self) -> DatabaseConfig:
        config = DatabaseConfig()
        _, selected_orm = pick(
            ['Sqlalchemy'],
            'Do you want use sqlalchemy or sqlmodel?',
            '=>',
        )
        config.orm = str(selected_orm).lower()
        _, use_alembic = pick(
            [
                'Yes, i want use alembic to manage migrations',
                "No, I don' want alembic",
            ],
            'Do you want use alembic to manage your database migrations?',
            '=>',
        )
        config.use_alembic = use_alembic == 0
        return config

    def setup_user_module(self):
        config = UserConfig()
        _, create_usermodel = pick(
            ['Yes, create user module', "I don' waht user module"],
            'Create a user module?',
            '=>',
        )
        if create_usermodel == 0:
            config.create_module = True
            _, jwt_store = pick(
                ['Headers', 'Cookies', 'Both'],
                'Where you want to get JWT token?',
                '=>',
            )
            config.store_jwt = str(jwt_store).lower()

            _, create_auth_system = pick(
                ['Yes, create auth system', 'No'],
                'Do you want create auth system?',
                '=>',
            )
            config.auth = create_auth_system == 0

        return config

    def setup_modules(self) -> ModuleConfig:
        config = ModuleConfig()
        path, _ = pick(
            ['modules/', 'modules/api/'], 'Where you want store modules?', '=>'
        )
        config.module_path = path
        return config


def cli():
    if len(sys.argv) < 3:
        print('Please tell a valid option ["new app", "new module"]')
    if sys.argv[1] == 'new':
        if sys.argv[2] == 'app':
            if len(sys.argv) == 4:

                bife = BifeCreateProject()
                bife.run()
            else:
                print('Error, tell a application name')
                sys.exit(1)
        elif sys.argv[2] == 'module':
            print('new module')



from fastapi import APIRouter, Depends, HTTPException, Response, status
from libs.security.security import Security

from .events import UserEvents
from .ext import (
    InvalidEmail,
    InvalidGender,
    InvalidName,
    InvalidPassword,
    InvalidUserLogin,
    InvalidUsername,
    UserAlreadyExists,
    UserNotFound,
)
from .models import UserModel
from .repository import UserRepository
from .schemas import LoginUserRequest, RegisterUserRequest, UserAuthResponse
from .service import UserService

router: APIRouter = APIRouter(tags=['user'], prefix='/user')
security: Security = Security()


class UserController:

    repository: UserRepository = UserRepository()
    service: UserService = UserService()
    events: UserEvents = UserEvents()

    @router.post('/login', response_model=UserAuthResponse)
    def user_login(self, data: LoginUserRequest, response: Response):
        try:
            user = self.repository.login_user(data)
        except UserNotFound as e:
            raise HTTPException(status.HTTP_404_NOT_FOUND, e.args[0])
        except (InvalidPassword, InvalidUserLogin) as e:
            raise HTTPException(status.HTTP_400_BAD_REQUEST, e.args[0])

        tokens = self.service.generate_access_token(user.uuid)
        response.set_cookie('access_token', tokens['access_token'])
        response.set_cookie(
            'refresh_access_token', tokens['refresh_access_token']
        )
        return UserAuthResponse(**tokens)

    @router.post('/register', response_model=UserAuthResponse)
    def register_user(self, data: RegisterUserRequest, response: Response):
        try:
            user = self.repository.register_user(data)
        except (
            InvalidUsername,
            InvalidEmail,
            InvalidGender,
            InvalidName,
            InvalidPassword,
        ) as e:
            raise HTTPException(status.HTTP_400_BAD_REQUEST, e.args[0])
        except UserAlreadyExists as e:
            raise HTTPException(status.HTTP_409_CONFLICT, e.args[0])
        tokens = self.service.generate_access_token(user.uuid)
        response.set_cookie('access_token', tokens['access_token'])
        response.set_cookie(
            'refresh_access_token', tokens['refresh_access_token']
        )
        return UserAuthResponse(**tokens)

    @router.get('/me')
    def get_about_user(
        self,
        current_user: UserModel = Depends(security.get_current_user),
    ):
        return current_user.__dict__



import uvicorn
from config import get_config
from fastapi import APIRouter, FastAPI, HTTPException, Request
from fastapi.responses import JSONResponse, RedirectResponse
from starlette.responses import StreamingResponse


class ApiHandler(FastAPI):

    config = get_config()

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def add_router(self, router: APIRouter) -> None:
        self.include_router(router)

    def start(self):

        uvicorn.run(
            'main:api_handler',
            host=self.config.API_HOST,
            port=self.config.API_PORT,
            reload=True,
        )



from fastapi import APIRouter, Depends
from libs.security.security import Security
from modules.user.models import UserModel

from .repository import TodoRepository
from .schemas import CreateTodoSchema, TodoUpdateSchema

router = APIRouter(prefix='/todo', tags=['todo'])
security = Security()


class TodoController:

    repository: TodoRepository = TodoRepository()

    @router.get('/')
    def get_all_todos(
        self, current_user: UserModel = Depends(security.get_current_user)
    ):
        self.repository.get_all_todos()
        return

    @router.post('/')
    def get_all_todos(
        self, current_user: UserModel = Depends(security.get_current_user)
    ):
        self.repository.new_todo(
            current_user.id,
        )
        return



from typing import Optional, TypedDict

from libs.database import DatabaseHandler
from libs.security.security import AccessTokenDict, Security
from sqlalchemy import or_, select
from utils import is_email_valid

from .ext import (
    InvalidEmail,
    InvalidGender,
    InvalidName,
    InvalidPassword,
    InvalidUsername,
)
from .models import UserModel
from .schemas import RegisterUserRequest


class UserInUseFields(TypedDict):
    username: bool
    email: bool


class UserService:

    database_handler = DatabaseHandler()
    security = Security()

    def validate_register_data(self, data: RegisterUserRequest) -> bool:
        # Valid name
        if len(data.fullname) <= 0:
            raise InvalidName('Name is empty')

        # Valid username
        if len(data.username) < 4:
            raise InvalidUsername('Username too short')
        if len(data.username) > 15:
            raise InvalidUsername('Username too long')

        # Email validate
        if not is_email_valid(data.email):
            raise InvalidEmail('Invalid email')

        # Password validate
        self.validate_password(data.password)

        # Gender validate
        if not self.get_gender_by_name(data.gender):
            raise InvalidGender('Gender not found')

        return True

    def validate_password(self, password: str) -> bool:
        if len(password) < 8:
            raise InvalidPassword('Password to short, min 8 characters')
        if len(password) > 25:
            raise InvalidPassword('Password to long, max 25 characters')

    def get_user_by_email_or_username(
        self, username: str = '', email: str = ''
    ) -> Optional[UserModel]:
        with self.database_handler as session:
            query = select(UserModel).where(
                or_(UserModel.email == email, UserModel.username == username)
            )
            user = session.scalar(query)
        return user

    def valid_email_or_username_in_use(
        self, data: RegisterUserRequest
    ) -> UserInUseFields:
        result: UserInUseFields = {'email': False, 'username': False}
        user = self.get_user_by_email_or_username(data.username, data.email)
        if user:
            result['email'] = user.email == data.email
            result['username'] = user.username == data.username
        return result

    def generate_access_token(self, sub) -> AccessTokenDict:
        return {
            'access_token': self.security.create_access_token(sub),
            'refresh_access_token': self.security.create_refresh_access_token(
                sub
            ),
        }



from dataclasses import dataclass, field
from typing import Literal


@dataclass
class DatabaseConfig:
    orm: Literal['sqlalchemy', 'sqlmodel'] = field(default='sqlalchemy')
    use_alembic: bool = field(default=True)


@dataclass
class UserConfig:
    create_module: bool = field(default=True)
    store_jwt: Literal['headers', 'cookie', 'both'] = field(default='headers')
    auth: bool = field(default=True)


@dataclass
class ModuleConfig:
    module_path: str = field(default='modules/')
